#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2012  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import errno
import json
import os
import re
import sqlite3
import xml.dom.minidom

import xdg.BaseDirectory



def make_parent_dirs(path, mode=0o777, exist_ok=True):
  """
  Recursively create all parent directories of the given path.
  """
  try:
    dname = os.path.dirname(path)
    if dname:
      os.makedirs(dname, mode=mode, exist_ok=exist_ok)
  except OSError as e:
    if e.errno != errno.EEXIST:
      raise e



class Table():
  """
  Base class for creating database tables from remote data.

  Subclasses must override TABLE, FIELDS, and db_retrieve.
  """

  # Database table name.
  TABLE = ''

  # Database field names and types.
  FIELDS = (
#    ('Foo', 'TEXT'),
#    ('Bar', 'TEXT'),
  )

  # Useful temporary views.
  VIEWS = {
#     'View Name' : 'SELECT query'
  }

  def __init__(self, database=None, database_parameters={}):
    """
    Initialize the object.

    database: The database to pass to the sqlite3.connect, or an existing
    connection.

    database_parameters: Additional parameters to pass to sqlite3.connect.

    The database table will be checked during initialization and rebuilt if
    necessary.
    """

    if isinstance(database, sqlite3.Connection):
      self.conn = database
    else:
      if not database:
        cache_dir = xdg.BaseDirectory.save_cache_path('NIST')
        database = os.path.join(cache_dir, 'data.sqlite3')

      if database != ':memory:':
        # Ensure that the parent directories exist.
        make_parent_dirs(database, exist_ok=True)

      self.conn = sqlite3.connect(database, **database_parameters)

    if not self.db_is_consistent():
      self.db_rebuild()
    self.db_create_views()



  ##############################################################################
  # Database methods
  ##############################################################################

  def db_is_consistent(self):
    """
    Check if the database table is consistent, i.e. that it exists and
    contains the expected columns.

    The data itself is not checked.

    Returns True or False. The database is not modified.
    """
    with self.conn as c:
      expected_columns = self.FIELDS
      columns = c.execute('PRAGMA table_info("%s")' % self.TABLE).fetchall()
      if len(expected_columns) == len(columns):
        for (name, typ), meta in zip(expected_columns, columns):
          if name != meta[1] or typ != meta[2]:
            return False
        else:
          return True
      else:
        return False



  def db_rebuild(self):
    """
    Rebuild the database table.
    """
    with self.conn as c:
      c.execute('DROP TABLE IF EXISTS "%s"' % self.TABLE)
      columns = ['"%s" %s' % (n,t) for n, t in self.FIELDS]
      c.execute('CREATE TABLE "%s" (%s)' % (self.TABLE, ','.join(columns)))
      qs = ','.join('?' for x in self.FIELDS)
      for row in self.db_retrieve():
        c.execute('INSERT INTO "%s" VALUES (%s)' % (self.TABLE, qs), row)
      c.execute('VACUUM')



  def db_retrieve(self):
    """
    Retrieve the data and return the values to insert into the table.

    The result should be an iterator that produces tuples of the expected size.

    This method must be overridden in subclasses for each table.
    """
    pass



  def db_create_views(self):
    """
    Create temporary views in the database.
    """
    if self.VIEWS:
      for name, select in self.VIEWS.items():
        self.conn.execute('CREATE TEMP VIEW IF NOT EXISTS "%s" AS %s' % (name, select))



  ##############################################################################
  # Accessibility
  ##############################################################################

  def __iter__(self):
    with self.conn as c:
      for r in c.execute('SELECT * FROM "%s"' % self.TABLE):
        yield r



  def __getitem__(self, key):
    with self.conn as c:
      return c.execute('SELECT * FROM "%s" WHERE "%s"=?' % (self.TABLE, self.FIELDS[0][0]), (key,)).fetchone()



  def search(self, terms, union=True):
    """
    Search first field in table.

    See search_by_field for more help.
    """
    return self.search_by_field(self.FIELDS[0][0], terms, union)



  def search_by_field(self, field, terms, union=True):
    """
    Search the table for records in the given field "LIKE" the search terms.

    union: if True then the union of all matches is returned, otherwise the
    intersection.

    Use "%" as a wildcard, e.g. "search('%foo%')".
    """
    if field not in [n for n,t in self.FIELDS]:
      raise ValueError('unrecognized field in table "%s": "%s"' % (self.TABLE, field))
    else:
      if isinstance(terms, str):
        terms = (terms, )
      else:
        terms = tuple(terms)
      if union:
        glue = 'OR'
      else:
        glue = 'AND'
      where = glue.join('"%s" LIKE ?' % field for x in terms)
      query = 'SELECT * FROM "%s" WHERE %s' % (self.TABLE, where)
      with self.conn as c:
        for r in c.execute(query, terms):
          yield r



  ##############################################################################
  # String Output
  ##############################################################################

  def to_json_str(self, *args, **kwargs):
    """
    Get a JSON representation of the table.

    All arguments are passed to json.dumps
    """
    obj = []
    for c in self:
      obj.append(dict( (k,v) for (k,x), v in zip(self.FIELDS, c) ))
    return json.dumps(obj, *args, **kwargs)



  def to_xml_str(self, *args, **kwargs):
    """
    Get an XML representation of the table.

    All arguments are passed to xml.dom.minidom.Document.toprettyxml
    """
    doc = xml.dom.minidom.Document()
    base = doc.createElement(self.TABLE.replace(' ', '_'))
    doc.appendChild(base)
    for c in self:
      entry = doc.createElement('entry')
      base.appendChild(entry)
      for (k,x), v in zip(self.FIELDS, c):
        field = doc.createElement(k)
        entry.appendChild(field)
        if v:
          v = str(v)
          text = doc.createTextNode(v)
          field.appendChild(text)
    x = doc.toprettyxml(*args, **kwargs)
    # Correct the textnode output bug.
    x = re.sub(r'>\s*\n\s*([^\s<])', r'>\1', x)
    x = re.sub(r'([^\s>])\s*\n\s*<', r'\1<', x)
    x = re.sub(r'>\s*\n\s*\n\s*<', '><', x)
    return x

