#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2012  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
NIST atomic weights and isotope data.

The data is retrieved from the NIST servers at

  http://www.nist.gov/pml/data/comp.cfm

and stored locally in an SQLite3 database. For an explanation of notes and other
data in the table, see

  http://www.nist.gov/pml/data/comp-notes.cfm

Note that the term "weight" is used in the NIST source files and has therefore
been used here too.
"""

import re
import sys
import urllib.request

from SciNum import Decimal, UncertainDecimal, UncertainDecimalWithUnits as UDwU
from SciNum.units import Definition, Units
from NIST.DB import Table



def retrieve_isotopes():
  """
  Retrieve isotope data.

  Each isotope is returned as a dictionary.
  """
  url = 'http://physics.nist.gov/cgi-bin/Compositions/stand_alone.pl?ele=&all=all&ascii=ascii2&isotype=all'
  with urllib.request.urlopen(url) as f:
    # From ('Content-Type', 'text/html; charset=ISO-8859-1') header.
    encoding = f.getheader('Content-Type').split('charset=', 1)[1]
    isotope = {}
    for line in f:
      line = line.decode(encoding)
      try:
        # "&nbsp;" shouldn't even be in a preformatted field.
        key, value = [x.replace('&nbsp;', ' ').strip() for x in line.split('=', 1)]
      except ValueError:
        continue
      isotope[key] = value

      # This marks the end of an isotope, so insert it into the database.
      if key == 'Notes':
        # This could be done more neatly with an OrderedDict, but explicitly
        # passing the keys ensures that an error will be thrown if the input data
        # format changes unexpectedly. Obviously additional fields would be missed
        # but nothing would break because of it.
        yield isotope
        isotope.clear()





def split(formula, as_tuple=False):
  """
  Split a formula into a dictionary of atomic symbols and the number
  of times each symbol occurred in the formula.

  as_tuple: If True, return a dictionary of atomic symbols for one molecule
  along with the stoichiometric coefficient, otherwise return a dictionary with
  all values multiplied by the coefficient.

  The parser is simplistic but adequate for basic formulae. Both round and
  square brackets are recognized, along with LaTeX-style subscripts.

  Examples:
    H2O, CH4, CH_4, CH_3COOH, C(CH_3)_4, K+[PtCl3(C2H4)]-, K+[PtCl_3(C_2H_4)]-

  InChi representations are also recognized.
  """

  # There's certainly a more efficient way to do this, but building the
  # string by "expanding" nested brackets was the most programmatically
  # simple way of doing it that I thought of at the time of writing.
  # Besides, I added this method mostly as an afterthought for testing.

  # TODO
  # Improve this if possible.

  formula = formula.strip()
  # Recognize InChi
  if '/' in formula:
    formula = formula.split('/')[1]
  m = re.search(r'^(\d+)(.*)', formula)
  if m:
    coeff = int(m.group(1))
    formula = m.group(2)
  else:
    coeff = 1
  symbol_regex = re.compile(r'([A-Z][a-z]*)(?:_?({\d+}|\d+))?(.*)')
  atoms = {}
  # Recognize brackets
  for left, right in (('(', ')'), ('[', ']')):
    r = formula.find(right)
    while r > -1:
      l = formula.rfind(left, 0, r)
      if l > -1:
        a = formula[:l]
        b = formula[l+1:r]
        c = formula[r+1:]
        m = re.search(r'_?({\d+}|\d+)(.*)', c)
        if m:
          subscript = int(m.group(1).strip('{}'))
          c = m.group(2)
        else:
          subscript = 1
        formula = a + b * subscript + c
        r = formula.find(right)
      else:
        raise ValueError("unmatched brackets")
  m = symbol_regex.search(formula)
  while m:
    symbol = m.group(1)
    if m.group(2):
      subscript = int(m.group(2).strip('{}'))
    else:
      subscript = 1
    m = symbol_regex.search(m.group(3))
    try:
      atoms[symbol] += subscript
    except KeyError:
      atoms[symbol] = subscript
  if as_tuple:
    return atoms, coeff
  if coeff > 1:
    for k, v in atoms.items():
      atoms[k] = coeff * v
  return atoms





def get_uncertain_decimal(value):
  """
  Convert string value to an UncertainDecimal
  """
  value = value.strip('[]#')
  if value[-1] != ')':
    return Decimal(value)
  else:
    return UncertainDecimal(value)




class Isotopes(Table):
  """
  Work with NIST isotope data.

  The data will be retrieved from NIST's server and cached in a local SQLite3
  database.
  """

  # Database table name.
  TABLE = 'Isotopes'

  # Database field names and types. The names correspond to the names in the
  # NIST ASCII data.
  FIELDS = (
    ('Atomic Number',          'INTEGER'),
    ('Atomic Symbol',          'TEXT'),
    ('Mass Number',            'INTEGER'),
    ('Relative Atomic Mass',   'TEXT'),
    ('Isotopic Composition',   'TEXT'),
    ('Standard Atomic Weight', 'TEXT'),
    ('Notes',                  'TEXT')
  )

  # Useful temporary views..
  VIEWS = {
    'Standard Atomic Weights' :
      'SELECT DISTINCT "Atomic Number", "Atomic Symbol", "Standard Atomic Weight" FROM "{}" WHERE "Atomic Symbol" NOT IN ("D", "T")'.format(TABLE)
  }

  # Scientific numbers that may include uncertainties.
  SCIENTIFIC_NUMBER_FIELDS = (
    'Relative Atomic Mass',
    'Isotopic Composition',
    'Standard Atomic Weight'
  )



  def db_retrieve(self):
    """
    Retrieve isotope data.
    """
    for isotope in retrieve_isotopes():
      yield(tuple(isotope[k] for k, v in self.FIELDS))



  def get_standard_weights(self, symbols=None):
    """
    Return an iterator over tuples of atomic symbols and standard atomic
    weights.
    """
    with self.conn as c:
      # Yield the rows to prevent returning a cursor object.
      if symbols:
        qs = ','.join('?' for x in symbols)
        for r in c.execute('SELECT * FROM "Standard Atomic Weights" WHERE "Atomic Symbol" IN ({})'.format(qs), symbols):
          yield r
      else:
        for r in c.execute('SELECT * FROM "Standard Atomic Weights"'):
          yield r



  def dict_symbol_to_weight(self):
    """
    Return a dictionary of symbols to standard weights.

    Weights are returned as UncertainDecimals.
    """
    return dict((s,get_uncertain_decimal(w)) for z,s,w in self.get_standard_weights())



  def dict_atomic_number_to_weight(self):
    """
    Return a dictionary of atomic numbers to standard weights.

    Weights are returned as UncertainDecimals.
    """
    return dict((z,get_uncertain_decimal(w)) for z,s,w in self.get_standard_weights())



  def dict_atomic_number_to_symbol(self):
    """
    Return a dictionary of atomic numbers to symbols.
    """
    return dict((z,s) for z,s,w in self.get_standard_weights())



  def dict_symbol_to_atomic_number(self):
    """
    Return a dictionary of atomic numbers to symbols.
    """
    return dict((s,z) for z,s,w in self.get_standard_weights())






class Query(object):
  """
  A conventient wrapper around the Isotopes tables for querying data.
  """
  def __init__(self, isotopes=None):
    if not isotopes:
      isotopes = Isotopes()
    self.isotopes = isotopes
    self.symbols_to_weights = None
    self.zs_to_weights = None
    self.zs_to_symbols = None
    self.symbols_to_zs = None



  def weigh(self, formula):
    """
    Return the standard weight of a formula.

    See split().
    """
    if not self.symbols_to_weights:
      self.symbols_to_weights = self.isotopes.dict_symbol_to_weight()
    try:
      return sum(self.symbols_to_weights[k] * v for k, v in split(formula).items())
    except KeyError as e:
      raise ValueError('unrecognized symbol {}'.format(e))



  def molecular_weight(self, formula, u=Definition('u', Units({'g':1,'mol':-1}), 1)):
    """
    Return an UncertainDecimalWithUnits representing the molecular weight of the
    given formula.

    u: a units.Definition of the atomic mass unit. The symbol is ignored. The
    default is g/mol.
    """
    if not self.symbols_to_weights:
      self.symbols_to_weights = self.isotopes.dict_symbol_to_weight()
    try:
      atoms, coeff = split(formula, as_tuple=True)
      weight = sum(self.symbols_to_weights[k] * v for k, v in atoms.items())
    except KeyError as e:
      raise ValueError('unrecognized symbol {}'.format(e))
    return UDwU(u.factor * weight, u.units)



  def symbol_to_weight(self, s):
    """
    Return the standard weight for the given symbol.
    """
    if not self.symbols_to_weights:
      self.symbols_to_weights = self.isotopes.dict_symbol_to_weight()
    return self.symbols_to_weights[s]



  def z_to_weight(self, z):
    """
    Return the stamdard atomic weight of the given atomic number.
    """
    if not self.zs_to_weights:
      self.zs_to_weights = self.isotopes.dict_atomic_number_to_weight()
    return self.zs_to_weights[z]



  def z_to_symbol(self, z):
    """
    Return the symbol corresponding to the given atomic number.
    """
    if not self.zs_to_symbols:
      self.zs_to_symbols = self.isotopes.dict_atomic_number_to_weight()
    return self.zs_to_symbols[z]



  def symbol_to_z(self, s):
    """
    Return the atomic number corresponding to the given symbol.
    """
    if not self.symbol_to_z:
      self.symbol_to_z = self.isotopes.dict_symbol_to_atomic_number()
    return self.symbol_to_z[s]




def parse_args(args=None):
  """
  Parse command-line arguments.

  If no arguments are passed then arguments are read from sys.argv.
  """
  import argparse
  parser = argparse.ArgumentParser(description='Display NIST atomic weight and isotope data. Source: http://www.nist.gov/pml/data/comp.cfm',
    epilog='Search terms may use "%" as wildcards, e.g. "%foo%" will match anything containing "foo". When "-w" is used, arguments are interpretted as chemical formulae or InChis and their atomic weight is calculated. If no arguments are passed then a table of atomic weights will be displayed. For more information such as an explanation of the notes, see http://www.nist.gov/pml/data/comp-notes.cfm')
  parser.add_argument('args', metavar='args', nargs='*', default=[])
  parser.add_argument('-z', '--number', dest='z', action='store_true',
    help='search by atomic number (the default)')
  parser.add_argument('-s', '--symbol', dest='symbol', action='store_true',
    help='search by atomic symbol')
  parser.add_argument('-m', '--mass', dest='mass', action='store_true',
    help='search by mass number')
  parser.add_argument('--union', dest='union', action='store_true',
    help='display the union of matching results instead of the intersection')
  parser.add_argument('-w', '--weight', dest='weight', action='store_true',
    help='calculate standard weights instead of searching for isotope data')

  group = parser.add_argument_group('display options')
  group.add_argument('--box', dest='box', action='store_true',
    help='render table with box-drawing characters')
  group.add_argument('--format', dest='format', action='store',
    default='g', metavar='<python format string>',
    help='a Python format string to use for numerical output')
  group.add_argument('--max-width', dest='max_width', action='store',
    type=int, default=25, metavar='n',
    help='maximum width of formula field when displaying weights')

  group = parser.add_argument_group('database options')
  group.add_argument('--database', dest='database', action='store', metavar='<path>',
    help='set the database path)')
  group.add_argument('--refresh', dest='refresh', action='store_true',
    help='refresh the data in the database (almost always unnecessary)')
  return parser.parse_args(args)



def main(args=None):
  try:
    from Tabulator import Tabulator, align_char, homogenize
  except ImportError:
    print("Install the Tabulator module (python3-tabulator) to use this.")
    return

  pargs = parse_args(args)

  isotopes = Isotopes(database=pargs.database)
  if pargs.refresh:
    isotopes.db_rebuild()
  query = Query(isotopes)

  if pargs.weight:
    args = pargs.args
    if args:
      # Print the arguments and their weights
      mw = pargs.max_width
      display_args = (x[:mw]+'…' if len(x) > mw else x for x in args)
      rows = []
      errors = []
      for formula, disp in zip(args, display_args):
        try:
          weight = query.weigh(formula)
        except ValueError as e:
          errors.append((formula, str(e)))
          continue
        # Some values may be certain.
        try:
          w, u = weight.format(pargs.format)
        except AttributeError:
          w = weight.__format__(pargs.format)
          u = ''
        rows.append((disp,w,u))
      rows = align_char(rows, cols=(1,2,3))
      rows = homogenize(rows, cols=(0,))
      headers = ('Formula', 'Weight', 'Uncertainty')
      rows = (headers,) + tuple(rows)
      tab = Tabulator(rows, cols=('c'), headers=('c'))
      if pargs.box:
        print(tab.to_box_table(left=True, top=True))
      else:
        print(tab)
      for e in errors:
        print('{}\n  {}'.format(*e))


    else:
      # Print the table of standard weights.
      headers = ('Z', 'Symbol', 'Standard Weight', 'Uncertainty')
      rows = []
      for z,s,w, in isotopes.get_standard_weights():
        z = str(z)
        ud = get_uncertain_decimal(w)
        # Some values may be certain, i.e. Decimals.
        try:
          w, u = ud.format(pargs.format)
        except AttributeError:
          w = ud.__format__(pargs.format)
          u = ''
        rows.append((z,s,w,u))
      rows = align_char(rows, cols=(0,2,3))
      rows = homogenize(rows, cols=(1,))
      rows = (headers,) + tuple(rows)
      tab = Tabulator(rows, cols=('c'), headers=('l', 'l', 'c'))
      if pargs.box:
        print(tab.to_box_table(left=True, top=True))
      else:
        print(tab)

  else:
    # Print out all of the isotopes
    headers = tuple(x[0] for x in isotopes.FIELDS)
    rows = []
    if pargs.args:
      if pargs.z:
        field = isotopes.FIELDS[0][0]
      elif pargs.symbol:
        field = isotopes.FIELDS[1][0]
      elif pargs.mass:
        field = isotopes.FIELDS[2][0]
      else:
        field = isotopes.FIELDS[0][0]
      for r in isotopes.search_by_field(field, pargs.args, pargs.union):
        rows.append(tuple(str(x) for x in r))
    else:
      for r in isotopes:
        rows.append(tuple(str(x) for x in r))
    rows = align_char(rows, cols=(3,4,5))
    rows = homogenize(rows, cols=(1,6))
    rows = homogenize(rows, cols=(0,2), align='r')
    rows = (headers,) + tuple(rows)
    tab = Tabulator(rows, cols=('c'), headers=('c'))
    if pargs.box:
      print(tab.to_box_table(left=True, top=True))
    else:
      print(tab)



if __name__ == '__main__':
  try:
    main(sys.argv[1:])
  except (KeyboardInterrupt, BrokenPipeError):
    pass
