#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2012  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""NIST physical constants.

The data is retrieved from the NIST servers at

  http://physics.nist.gov/cuu/Constants/

and stored locally in an SQLite3 database.
"""

import re
import sys
import urllib.request

from NIST.DB import Table
from SciNum.decimal import Decimal, pi
from SciNum import UncertainDecimalWithUnits as UDwU, MeasuredDecimalWithUnits as MDwU, DecimalWithUnits as DwU

def nist_retrieve_ascii_table():
  """
  Retrieve the NIST ASCII table of physical constants.
  """

  url = 'http://physics.nist.gov/cuu/Constants/Table/allascii.txt'
  with urllib.request.urlopen(url) as f:
    return f.read().decode('utf-8')


# This would be easier if the ascii file were regularly formatted.
def nist_parse_ascii_table(ascii):
  """
  Parse the NIST ASCII table of physical constants

  Returns an iterator over tuples of (quantity, value, uncertainty, unit).
  """

  ascii = re.sub(r'(\d\d) (\d\d)', r'\1\2', ascii)
  ascii = re.sub(r'(\d\d) (\d\d)', r'\1\2', ascii)
  ascii = re.sub(r'(?<=\d) e(?=-\d)', r'e', ascii)
  ascii = re.sub(r'(?<=\d) e(?=\d)', r'e', ascii)
  ascii = re.sub(r'(?<=\d)(\.\.\.) e(?=-\d)', r'\1e', ascii)
  ascii = re.sub(r'(?<=\d)(\.\.\.) e(?=\d)', r'\1e', ascii)
  ascii = re.sub(r'[\t ]+', r' ', ascii)

  for line in ascii.split('\n'):
    line = line.strip()
    if not line:
      continue
    m = re.search(r'^(.+?) (-?\d\S*) (0.\d\S*|\(exact\))(?: (.*))?$', line)
    if m:
      quantity, value, uncertainty, unit = m.groups()
      quantity = quantity.replace('mag.', 'magnetic')
      quantity = quantity.replace('mom.', 'moment')
      if uncertainty == '(exact)':
        uncertainty = None
      yield((quantity, value, uncertainty, unit))
    elif re.search(r'\d', line):
      print("warning: failed to parse line in physical constants table\n\t", line)






class PhysicalConstants(Table):
  """
  Work with NIST Physical Constant data.

  The data will be retrieved from NIST's server and cached in a local SQLite3
  database.
  """

  # Database table name.
  TABLE = 'Physical Constants'

  # Database field names and types. The names correspond to the names in the
  # NIST ASCII data.
  FIELDS = (
    ('Quantity',    'TEXT'),
    ('Value',       'TEXT'),
    ('Uncertainty', 'TEXT'),
    ('Unit',        'TEXT'),
  )



  def db_retrieve(self):
    """
    Retrieve physical constant date from NIST
    """
    return nist_parse_ascii_table(nist_retrieve_ascii_table())



  def __getitem__(self, key):
    with self.conn as c:
      r = c.execute('SELECT * FROM "{}" WHERE "{}"=?'.format(self.TABLE, self.FIELDS[0][0]), (key,)).fetchone()
      if r:
        q, v, dv, u = r
        if not dv:
          if '...' in v:
            if q == 'electric constant':
              m = self['magnetic constant']
              c = self['speed of light in vacuum']
              e = (1 / (m * c**2))._value
              return MDwU(e, u)
            elif q == 'hertz-inverse meter relationship':
              c = self['speed of light in vacuum']
              e = (1 / c**2)._value
              return MDwU(e, u)
            elif q == 'kilogram-joule relationship':
              c = self['speed of light in vacuum']
              e = (c**2)._value
              return MDwU(e, u)
            elif q == 'magnetic constant':
              v = (4 * pi()) / 10**7
              return MDwU(v, u)
            v = v.replace('...', '')
            return MDwU(v, u).to_uncertain_decimal()
          else:
            return DwU(v, u)
        else:
          return UDwU(v, u, plus=dv)
      else:
        return None



#   def exact_values(self):
#     with self.conn as c:
#       return c.execute('SELECT * FROM "{}" WHERE "Uncertainty" IS NULL'.format(self.TABLE,))





def parse_args(args=None):
  """
  Parse command-line arguments.

  If no arguments are passed then arguments are read from sys.argv.
  """
  import argparse
  parser = argparse.ArgumentParser(description='Query NIST physical constants. Source: http://physics.nist.gov/cuu/Constants/',
    epilog='Search terms may use "%" as wildcards, e.g. "%foo%" will match anything containing "foo". If no search terms are given then all constants will be returned.')
  parser.add_argument('terms', metavar='<search term>', nargs='*')
  parser.add_argument('--union', dest='union', action='store_true',
    help='display the union of matching results instead of the intersection')

  group = parser.add_argument_group('display options')
  group.add_argument('--box', dest='box', action='store_true',
    help='render table with box-drawing characters')
  group.add_argument('--max-width', dest='max_width', action='store',
    type=int, default=0, metavar='n',
    help='maximum width of formula field when displaying weights')

  group = parser.add_argument_group('database options')
  group.add_argument('--database', dest='database', action='store', metavar='<path>',
    help='set the database path')
  group.add_argument('--refresh', dest='refresh', action='store_true',
    help='refresh the data in the database (almost always unnecessary)')
  return parser.parse_args(args)




def main(args=None):
  try:
    from Tabulator import Tabulator, align_char, homogenize
  except ImportError:
    print("Install the Tabulator module (python3-tabulator) to use this.")
    return

  pargs = parse_args(args)

  pc = PhysicalConstants(database=pargs.database)
  if pargs.refresh:
    pc.db_rebuild()

  if pargs.terms:
    raw_rows = pc.search(pargs.terms, pargs.union)
  else:
    raw_rows = pc
  rows = []
  mw = pargs.max_width
  for r in raw_rows:
    rows.append(['' if x == None else x for x in r])
    if mw > 0 and len(rows[-1][0]) > mw:
      rows[-1][0] = rows[-1][0][:mw] + '…'
  for r in rows:
    for i in (1,2):
      v = r[i]
      if v:
        if '...'  in v:
          v = v.replace('...', '')
          v = Decimal(v).__format__('e').replace('e', '...e')
        else:
          v = Decimal(v).__format__('e')
        # Alignment hack
        if not '.' in v:
          if 'e' in v:
            v = v.replace('e', '.e')
          else:
            v += '.'
        r[i] = v

  headers = (n for n,t in pc.FIELDS)
  if rows:
    rows = align_char(rows, cols=(1,2))
    rows = homogenize(rows)
    rows = (headers,) + tuple(rows)
    tab = Tabulator(rows, cols=('c'), headers=('c'))
    if pargs.box:
      tab = tab.to_box_table(left=True, top=True)
    print(tab)


if __name__ == '__main__':
  try:
    main()
  except (KeyboardInterrupt, BrokenPipeError):
    pass